/******************************************************************************
  lastMAC.c -- els lastMAC functions
  
  Author: Mark Baranowski
  Email:  requestXXX@els-software.org (remove XXX)
  Download: http://els-software.org

  Last significant change: January 29, 2019
  
  These functions are provided "as is" in the hopes that they might serve some
  higher purpose.  If you want these functions to serve some lower purpose,
  then that too is all right.  So as to keep these hopes alive you may
  freely distribute these functions so long as this header remains intact.
  You may also freely distribute modified versions of these functions so long
  as you indicate that such versions are modified and so long as you
  provide access to the unmodified original copy.
  
  Note: The most recent version of these functions may be obtained from
        http://els-software.org

  ****************************************************************************/


#include "sysdefs.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>	/* sys/stat.h needed by els.h */

#include "defs.h"
#include "els.h"
#include "lastMAC.h"

/* Some ELS searches can run for days, so allow a week's worth of
   time-jitter before assuming a file's date unlikely: */
#define UNLIKELY_TIME(ftime) ((ELS_time_t)(ftime) > (ELS_time_t)The_Time_in_a_week)

Local void lastMAC_warn(Boole *warnedOnce, char *msg,
			Dir_Item *file, time_t ftime, char *dname);

Local ELS_time_t lastDir = 0;
Local ELS_time_t lastMod = 0;
Local ELS_time_t lastAcc = 0;
Local ELS_time_t lastChg = 0;


void lastMAC(Dir_Item *file, char *dname)
{
  /* Latest mod time of directories, e.g. "+GmN +TF +F'T{d}'": */
  if (file->isdir)
  {
    if (UNLIKELY_TIME(file->info.st_mtime))
    {
      static Boole warnedOnce = FALSE;
      lastMAC_warn(&warnedOnce, "directory modification",
		   file, file->info.st_mtime, dname);
    }
    else if ((ELS_time_t)file->info.st_mtime > lastDir)
      lastDir = (ELS_time_t)file->info.st_mtime;
  }

  if (!file->isdir && !file->islink)
  {
    /* Latest mod time except directories and symlinks, e.g. "+M +F'T{~l}'": */
    if (UNLIKELY_TIME(file->info.st_mtime))
    {
      static Boole warnedOnce = FALSE;
      lastMAC_warn(&warnedOnce, "file modification",
		   file, file->info.st_mtime, dname);
    }
    else if ((ELS_time_t)file->info.st_mtime > lastMod)
      lastMod = (ELS_time_t)file->info.st_mtime;

    /* Latest access time except directories and symlinks, e.g. "+MM": */
    if (UNLIKELY_TIME(file->info.st_atime))
    {
      static Boole warnedOnce = FALSE;
      lastMAC_warn(&warnedOnce, "file access",
		   file, file->info.st_atime, dname);
    }
    else if ((ELS_time_t)file->info.st_atime > lastAcc)
      lastAcc = (ELS_time_t)file->info.st_atime;
  }

  /* Latest change time of all files, e.g. "+MMM": */
  if (UNLIKELY_TIME(file->info.st_ctime))
  {
    static Boole warnedOnce = FALSE;
    lastMAC_warn(&warnedOnce, "change",
		 file, file->info.st_ctime, dname);
  }
  else if ((ELS_time_t)file->info.st_ctime > lastChg)
    lastChg = (ELS_time_t)file->info.st_ctime;

  return;
}


Local void lastMAC_warn(Boole *warnedOnce, char *msg,
			Dir_Item *file, time_t ftime, char *dname)
{
  if (*warnedOnce) return;
  *warnedOnce = TRUE;
  if (!warning_suppress)
  {
    char buff[128];
    T_print(buff, "F", ftime, NULL, FALSE, FALSE);
    fprintf(stderr, "%s: Improbable %s time: %s %s\n",
	    Progname, msg, buff, full_name(dname, file->fname));
    fprintf(stderr, "%s: Ignoring subsequent %s time warnings...\n",
	    Progname, msg);
  }
  /*listingError = TRUE; -- FIXME, is this a desired action? */
  return;
}


void lastMAC_print(void)
{
  char buff[128];

  T_print(buff, "F", lastDir, NULL, FALSE, FALSE);
  printf("LastModDir:  %s\n", buff);
  T_print(buff, "F", lastMod, NULL, FALSE, FALSE);
  printf("LastModFile: %s\n", buff);
  T_print(buff, "F", lastAcc, NULL, FALSE, FALSE);
  printf("LastAccess:  %s\n", buff);
  T_print(buff, "F", lastChg, NULL, FALSE, FALSE);
  printf("LastChange:  %s\n", buff);
  printf("\n");

  return;
}


void lastMAC_print2(void)
{
  char buff[256];
  char *bp;

  bp = buff;
  bp = T_print(bp, "F",  lastDir, NULL, FALSE, FALSE);
  printf("LastDir: %s\n", buff);
  bp = buff;
  bp = T_print(bp, "F ", lastMod, NULL, FALSE, FALSE);
  bp = T_print(bp, "F ", lastAcc, NULL, FALSE, FALSE);
  bp = T_print(bp, "F" , lastChg, NULL, FALSE, FALSE);
  printf("LastMAC: %s\n", buff);
  printf("\n");

  return;
}
